//
// Copyright (c) 2014, Peter A.M. van Hoof.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// Peter A.M. van Hoof
// Royal Observatory of Belgium
// Ringlaan 3
// B-1180 Brussels
// Belgium
// p.vanhoof@oma.be
//
//
// If you use any of these data in a scientific publication, please refer to
//
// van Hoof P.A.M., Williams R.J.R., Volk K., Chatzikos M., Ferland G.J., Lykins M., Porter R.L.
// Accurate determination of the free-free Gaunt factor, II -- relativistic Gaunt factors
// 2015, MNRAS, 449, 2112
//

#include "mpreal.h"
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <complex>
#include <vector>
#include <algorithm>
#include <mpi.h>
#include <cassert>
#include <fstream>
#include <map>
#include <unistd.h>
#include <sys/file.h>

using namespace std;
using namespace mpfr;

static const mp_prec_t prec_start = 256;

template<typename T>
T pow2(T x) { return x*x; }

inline mpreal operator "" _mp( unsigned long long l )
{
	return mpreal(size_t(l));
}
inline mpreal operator "" _mp( long double l )
{
	return mpreal(l);
}
inline mpreal operator "" _mp( const char* s, size_t )
{
	return mpreal(s);
}

class loss_of_precision
{
public:
	loss_of_precision() {}
};

mpreal g_iii_norm_rel(const mpreal&,const mpreal&,mpreal&,const mpreal&);
mpreal g_iii_av_rel(const mpreal&,const mpreal&,const mpreal&,mpreal&,const mpreal&);
void gaunt_step(const mpreal[],mpreal(*)(const mpreal&,const mpreal[],mpreal&,mp_prec_t&),
		mpreal&,mpreal&,const mpreal&,mpreal&,mpreal&,int&);
mpreal g_fun1(const mpreal&,const mpreal[],mpreal&,mp_prec_t&);
mpreal g_fun2(const mpreal&,const mpreal[],mpreal&,mp_prec_t&);

unsigned int nCPU;
unsigned int nRANK;

int main(int argc, char* argv[])
{
	mpreal::set_default_prec(prec_start);

	MPI_Init( &argc, &argv );
	int arg;
	MPI_Comm_size( MPI_COMM_WORLD, &arg );
	nCPU = arg;
	MPI_Comm_rank( MPI_COMM_WORLD, &arg );
	nRANK = arg;

	unsigned int g2max = 81;
	unsigned int umax = 146;

	mpreal g2start = -6_mp;
	mpreal ustart = -16_mp;
	mpreal steps_per_dex = 5_mp;

	vector<double> resv( g2max*umax );
	vector<double> dresv( g2max*umax );

	for( int ipz = 1; ipz <= 36; ++ipz ) {

		for( unsigned int i=0; i < resv.size(); ++i ) {
			resv[i] = 0.;
			dresv[i] = 0.;
		}

		mpreal Z = ipz;

		if( nRANK == 0 ) {
			// this rank does not do any work, it only sends out jobs to the other ranks
			vector<int> jobs( g2max*umax + nCPU-1 );
			unsigned int k = 0;
			// jobs for high gamma^2 tend to be longer running, so we do those first
			for( int g2=g2max-1; g2 >= 0; --g2 )
				for( unsigned int u=0; u < umax; ++u )
					jobs[k++] = u*g2max + g2;
			// this is the end-of-jobs signal for the other ranks
			for( unsigned int i=1; i < nCPU; ++i )
				jobs[k++] = -1;

			assert( k == jobs.size() );

			for( unsigned int k=0; k < jobs.size(); ++k ) {
				int dest = -1;
				MPI_Status status;
				MPI_Recv( &dest, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
				int job = jobs[k];
				MPI_Send( &job, 1, MPI_INT, dest, 0, MPI_COMM_WORLD );
			}
		}
		else {
			// these ranks do the actual work
			vector<mpreal> norm( g2max );
			vector<mpreal> rnorm( g2max );

			// set up normalization factors first, these do not depend on u
			for( unsigned int g2=0; g2 < g2max; ++g2 ) {
				mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
				mpreal rerr, rerr2;
				mpreal res = g_iii_norm_rel( exp10(gam2), Z, rerr, 1.e-5_mp );
				mpreal res2 = g_iii_norm_rel( exp10(gam2), Z, rerr2, 1.e-7_mp );
				assert( abs(res/res2-1_mp)/rerr < 2_mp );
				assert( rerr < 2e-5_mp );
				norm[g2] = res2;
				rnorm[g2] = rerr2;
			}

			while( true ) {
				MPI_Status status;
				MPI_Send( &nRANK, 1, MPI_UNSIGNED, 0, 0, MPI_COMM_WORLD );
				int job;
				MPI_Recv( &job, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
				if( job < 0 )
					break;
				int g2 = job%g2max;
				int u = job/g2max;
				mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
				mpreal uu = mpreal(u)/steps_per_dex + ustart;
				mpreal res2, rerr, rerr2;
				mpreal tol = 1.e-5_mp;
				mpreal res = g_iii_av_rel( exp10(gam2), exp10(uu), Z, rerr, tol );
				for( int k=0; k < 3; ++k ) {
					res2 = g_iii_av_rel( exp10(gam2), exp10(uu), Z, rerr2, tol/10_mp );
					if( abs(res/res2-1_mp)/rerr >= 3_mp ) {
						res = res2;
						rerr = rerr2;
						tol /= 10_mp;
					}
					else {
						break;
					}
				}
				if( abs(res/res2-1_mp)/rerr >= 3_mp ) {
					cerr << "FAILURE " << gam2 << " " << uu << " " << Z << endl;
					throw loss_of_precision();
				}
				assert( rerr < 3e-5_mp );
				res2 /= norm[g2];
				mpreal err = hypot(rerr2,rnorm[g2])*res2;
				cerr << nRANK << " " << exp10(gam2) << " " << exp10(uu) << " ";
				cerr << res << " " << err << endl;
				resv[ job ] = double(res2);
				dresv[ job ] = double(err);
			}
		}

		MPI_Barrier( MPI_COMM_WORLD );

		vector<double> resc = resv;
		MPI_Reduce( &resc[0], &resv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
		vector<double> dresc = dresv;
		MPI_Reduce( &dresc[0], &dresv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

		MPI_Barrier( MPI_COMM_WORLD );

		if( nRANK == 0 ) {
			ostringstream oss;
			oss << "gauntff_bhe_Z" << setw(2) << setfill( '0' ) << ipz << ".dat";
			cerr << "writing file " << oss.str() << endl;
			ofstream io( oss.str().c_str() );

			io << setw(9) << 20140510 << " # magic number" << endl;
			io << setw(4) << g2max << " " << setw(4) << umax;
			io << " # number of grid points in gam2 and u direction" << endl;
			io << setw(9) << g2start << " # start value for log(gam2)" << endl;
			io << setw(9) << ustart << " # start value for log(u)" << endl;
			io << setw(9) << 1_mp/steps_per_dex << " # stepsize in dex for log(gam2) and log(u)" << endl;
			io << setw(9) << Z << " # atomic number Z" << endl;
			io << "#" << endl;
			io << "# thermally averaged relativistic Gaunt factors. Rows have constant log(u),";
			io << " columns have constant log(gam2)" << endl;
			for( unsigned int k=0; k < resv.size(); ++k ) {
				io << scientific << setprecision(8) << setw(16) << resv[k];
				if( (k+1)%g2max == 0 )
					io << endl;
			}
			io << "#" << endl;
			io << "# uncertainties in the thermally averaged relativistic Gaunt factors" << endl;
			for( unsigned int k=0; k < resv.size(); ++k ) {
				io << scientific << setprecision(4) << setw(12) << dresv[k];
				if( (k+1)%g2max == 0 )
					io << endl;
			}

			io.close();
		}

		MPI_Barrier( MPI_COMM_WORLD );
	}

	cerr << nRANK << " done." << endl;

	MPI_Finalize();

	return 0;
}

mpreal g_iii_norm_rel(const mpreal& gam2, const mpreal& Z, mpreal& rerr, const mpreal& tol)
{
	const mpreal mc2 = 8.18710506e-14_mp;  // m_e*c^2 in J
	const mpreal Ryd = 2.179872171e-18_mp; // Ryd in J
	mpreal pi = 2_mp*asin(1_mp);
	mpreal lambda = pow2(Z)*Ryd/(gam2*mc2);
	const mpreal eta = -70_mp;
	mpreal nu = eta + 1_mp/lambda;

	mpreal par[4];
	par[0] = lambda;
	par[1] = nu;
	par[3] = Z;

	mpreal x = 1_mp/lambda;
	mpreal h = x*tol*min(1e-5_mp,1e-2_mp/pow2(gam2));
	// this is analytic approximation for integral from x to x+h
	// we do the first step analytically because the derivative of the integrand is infinite at x = 1/lambda
	mpreal res = (h/lambda*asinh(h/sqrt(2_mp*h/lambda)) + h*sqrt(2_mp*h/lambda+pow2(h))/2_mp)/
		(lambda*(1_mp + exp(1_mp/lambda-nu)));
	x += h;
	mpreal err = 0_mp;
	// error == 0 : integration needs more steps
	// error == 1 : normal termination of integral
	// error == 2 : step failed to converge
	int error = 0;
	do {
		gaunt_step( par, g_fun2, x, h, tol/2_mp, res, err, error );
	} while( error == 0 );

	if( error < 2 ) {
		rerr = err/res;
		return res;
	}
	else {
		cerr << "integration 1 failed: gam2 = " << gam2 << endl;
		throw loss_of_precision();
	}
}

mpreal g_iii_av_rel(const mpreal& gam2, const mpreal& u, const mpreal& Z, mpreal& rerr, const mpreal& tol)
{
	const mpreal mc2 = 8.18710506e-14_mp;  // m_e*c^2 in J
	const mpreal Ryd = 2.179872171e-18_mp; // Ryd in J
	mpreal pi = 2_mp*asin(1_mp);
	mpreal lambda = pow2(Z)*Ryd/(gam2*mc2);
	const mpreal eta = -70_mp;
	mpreal nu = eta + 1_mp/lambda;

	mpreal par[4];
	par[0] = lambda;
	par[1] = nu;
	par[2] = u;
	par[3] = Z;

	mpreal x = u + 1_mp/lambda;
	mpreal h = 1000_mp*tol*min(min(1_mp,u),gam2);
	mpreal res = 0_mp;
	mpreal err = 0_mp;
	int error = 0;
	do {
		gaunt_step( par, g_fun1, x, h, tol/2_mp, res, err, error );
	} while( error == 0 );

	if( error < 2 ) {
		// this is Eq. 24 of Nozawa, Itoh & Kohyama, 1998, ApJ, 507, 530
		// the factor exp(u) is already included in res
		// the normalization is done by the caller
		// the factor lambda^3 is already included in the norm used by the caller
		rerr = err/res;
		return 3_mp/32_mp*sqrt(6_mp*lambda/pi)*res;
	}
	else {
		cerr << "integration 2 failed: gam2 = " << gam2 << " u = " << u << endl;
		throw loss_of_precision();
	}
}

void gaunt_step(const mpreal par[],
		mpreal (*fun)(const mpreal&,const mpreal[],mpreal&,mp_prec_t&),
		mpreal& x,
		mpreal& h,
		const mpreal& toler,
		mpreal& yval,
		mpreal& dyval,
		int& error)
{
	if( error > 0 )
		return;

	const int MAX_ITER = 50;
	const int MAX_DIVS = 4;
	const int MAX_ABS = (1<<MAX_DIVS) + 1;

	mpreal y1[MAX_DIVS+1];
	mpreal dy1[MAX_DIVS+1];
	mpreal y1a[MAX_ABS];
	mpreal y1b[MAX_ABS];

	for( int k=0; k < MAX_ITER; ++k )
	{
		for( int i=0; i <= MAX_DIVS; ++i )
		{
			int nabs = 1<<i;
			int stride = 1<<(MAX_DIVS-i);
			mpreal h1 = h/mpreal(nabs);
			// for k > 0 the points at 0, MAX_DIVS/2 and MAX_DIVS are already set up...
			if( k == 0 || i > 1 )
			{
				for( int j=1; j <= nabs; ++j )
				{
					int index = j*stride;
					// y1a[index] may already have been initialized on a previous
					// iteration of the loop over i. If so, do not repeat the calculation
					if( i == 0 || j%2 == 1 )
					{
						mpreal x1 = x + h1*mpreal(j);
						mpreal rerr;
						mp_prec_t myprec;
						y1a[index] = (*fun)(x1,par,rerr,myprec);
						y1b[index] = rerr*y1a[index];
					}
				}
			}
			if( i < 2 ) {
				y1[i] = 0_mp;
				dy1[i] = 0_mp;
			}
			else {
				// use Eq. 4.1.20 from Numerical Recipes in C, 1st ed.
				y1[i] = (23_mp*y1a[stride]+7_mp*y1a[2*stride]+
					 13_mp*y1a[(nabs-1)*stride]+5_mp*y1a[nabs*stride])/12_mp;
				for( int j=3; j < nabs-1; ++j )
					y1[i] += y1a[j*stride];
				y1[i] *= h1;
				// this is the regular error propagation formula assuming errors are uncorrelated
				dy1[i] = (pow2(23_mp*y1b[stride])+pow2(7_mp*y1b[2*stride])+
					  pow2(13_mp*y1b[(nabs-1)*stride])+pow2(5_mp*y1b[nabs*stride]))/144_mp;
				for( int j=3; j < nabs-1; ++j )
					dy1[i] += pow2(y1b[j*stride]);
				dy1[i] = sqrt(dy1[i])*h1;
			}
			// check for convergence by comparing the integrals with stepsizes h1 and 2*h1
			if( i > 2 )
			{
				// toler is relative precision
				mpreal err = abs((y1[i]-y1[i-1])/((yval+y1[i])*toler));
				if( err <= 1_mp )
				{
					mpreal err = hypot(dy1[i],abs(y1[i]-y1[i-1])/2_mp);
					yval += y1[i];
					dyval = hypot(dyval,err);
					mpreal fac = 0.9_mp*cbrt(1_mp/max(err,0.1_mp));
					x += h;
					h = 8_mp*h1*fac;
					// make estimate of the remainder of the integral to infinity
					// assuming that g_iii() is constant
					// use additional factor 100 to make sure that the error from
					// truncating the integral is negligible compared to the other
					// terms and our error estimate is valid
					error = ( abs(y1[i]/(yval*expm1(h))) < toler/100_mp ) ? 1 : 0;
					return;
				}
			}
		}
		// convergence failed. Here we reduce the stepsize and try again...
		y1a[(MAX_ABS-1)/2] = y1a[1];
		y1a[MAX_ABS-1] = y1a[2];
		y1b[(MAX_ABS-1)/2] = y1b[1];
		y1b[MAX_ABS-1] = y1b[2];
		h /= mpreal((MAX_ABS-1)/2);
	}
	error = 2;
	return;
}

mpreal g_fun1(const mpreal& x, const mpreal par[], mpreal& rerr, mp_prec_t& myprec)
{
	// this routine implements Eqs. 16 through 19 of Nozawa, Itoh & Kohyama, 1998, ApJ, 507, 530
	// the integrand is scaled with a factor exp(u) in the first term of i1 to prevent underflow
	mpreal lambda = par[0];
	mpreal nu = par[1];
	mpreal u = par[2];
	mpreal Z = par[3];
	const mpreal alpha = 7.2973525698e-3_mp; // fine-structure constant
	mpreal pi = 2_mp*asin(1_mp);
	mpreal xml2 = pow2(x) - 1_mp/pow2(lambda);
	mpreal xml = sqrt(xml2);
	mpreal xml3 = xml*xml2;
	mpreal xuml2 = pow2(x-u) - 1_mp/pow2(lambda);
	mpreal xuml = sqrt(xuml2);
	mpreal xuml3 = xuml*xuml2;
	mpreal betaf = 2_mp*log((x-u+xuml)*lambda);
	mpreal betai = 2_mp*log((x+xml)*lambda);
	mpreal L = 2_mp*log(((x-u)*x + xuml*xml - 1_mp/pow2(lambda))*lambda/u);
	mpreal i1 = xml2/(exp(x-u-nu)+exp(-u))*(x-u)/x*(1_mp - 1_mp/(exp(x-u-nu)+1_mp));
	mpreal i2 = (1_mp - exp(-2_mp*pi*alpha*Z*x/xml))/(1_mp - exp(-2_mp*pi*alpha*Z*(x-u)/xuml));
	mpreal i3a = 4_mp/3_mp - 2_mp*(x-u)*x*(xuml2+xml2)/(xuml2*xml2);
	mpreal i3b = betaf*x/xuml3 + betai*(x-u)/xml3 - betaf*betai/(xuml*xml);
	mpreal i4a = 8_mp/3_mp*(x-u)*x/(xuml*xml);
	mpreal i4b = pow2(u)/(xuml3*xml3)*(pow2(x-u)*pow2(x) + xuml2*xml2);
	mpreal i4c = u/(2_mp*xuml*xml*pow2(lambda));
	mpreal i5a = ((x-u)*x + xml2)/xml3*betai;
	mpreal i5b = ((x-u)*x + xuml2)/xuml3*betaf;
	mpreal i5c = 2_mp*u*(x-u)*x/(xuml2*xml2);
	myprec = mpreal::default_prec;
	rerr = machine_epsilon(mpreal::default_prec);
	return i1*i2*(i3a + i3b/pow2(lambda) + L*(i4a + i4b + i4c*(i5a - i5b + i5c)));
}

mpreal g_fun2(const mpreal& x, const mpreal par[], mpreal& rerr, mp_prec_t& myprec)
{
	// this routine implements Eq. 10 of Nozawa, Itoh & Kohyama, 1998, ApJ, 507, 530
	// the integrand is scaled with a factor lambda^-3
	mpreal lambda = par[0];
	mpreal nu = par[1];
	mpreal xml2 = pow2(x) - 1_mp/pow2(lambda);
	myprec = mpreal::default_prec;
	rerr = machine_epsilon(mpreal::default_prec);
	return x*sqrt(xml2)/(exp(x-nu)+1_mp);
}
